@@ -21,7 +21,6 @@ import six
 import webob
 
 from placement import exception
-from placement.i18n import _
 from placement import lib
 from placement import microversion
 from placement.objects import allocation_candidate as ac_obj
@@ -30,8 +29,14 @@ from placement.schemas import allocation_candidate as schema
 from placement import util
 from placement import wsgi_wrapper
 
+# The microversions at which the schema used to validate
+# query parameters to GET /allocation_candidates differs.
+_GET_SCHEMA_MICROVERSIONS = [
+    (1, 36), (1, 35), (1, 33), (1, 31), (1, 25), (1, 21), (1, 17), (1, 16)
+]
 
-def _transform_allocation_requests_dict(alloc_reqs):
+
+def _transform_allocation_requests_dict(alloc_reqs, want_version):
     """Turn supplied list of AllocationRequest objects into a list of
     allocations dicts keyed by resource provider uuid of resources involved
     in the allocation request. The returned results are intended to be used
@@ -54,6 +59,12 @@ def _transform_allocation_requests_dict(alloc_reqs):
                     }
                 }
             },
+            # If microversion >=1.34 then map suffixes to providers.
+            "mappings": {
+                "_COMPUTE": [$rp_uuid2],
+                "": [$rp_uuid1]
+
+            },
         },
         ...
     ]
@@ -66,7 +77,10 @@ def _transform_allocation_requests_dict(alloc_reqs):
         for rr in ar.resource_requests:
             res_dict = rp_resources[rr.resource_provider.uuid]['resources']
             res_dict[rr.resource_class] = rr.amount
-        results.append(dict(allocations=rp_resources))
+        result = dict(allocations=rp_resources)
+        if want_version.matches((1, 34)):
+            result['mappings'] = ar.mappings
+        results.append(result)
 
     return results
 
@@ -193,8 +207,7 @@ def _transform_provider_summaries(p_sums, requests, want_version):
         ret[ps.resource_provider.uuid] = {'resources': resources}
 
         if include_traits:
-            ret[ps.resource_provider.uuid]['traits'] = [
-                t.name for t in ps.traits]
+            ret[ps.resource_provider.uuid]['traits'] = ps.traits
 
         if enable_nested_providers:
             ret[ps.resource_provider.uuid]['parent_provider_uuid'] = (
@@ -216,7 +229,7 @@ def _transform_allocation_candidates(alloc_cands, requests, want_version):
     """
     if want_version.matches((1, 12)):
         a_reqs = _transform_allocation_requests_dict(
-            alloc_cands.allocation_requests)
+            alloc_cands.allocation_requests, want_version)
     else:
         a_reqs = _transform_allocation_requests_list(
             alloc_cands.allocation_requests)
@@ -230,6 +243,17 @@ def _transform_allocation_candidates(alloc_cands, requests, want_version):
     }
 
 
+def _get_schema(want_version):
+    """Calculate the desired query parameter schema for
+    list_allocation_candidates.
+    """
+    for maj, min in _GET_SCHEMA_MICROVERSIONS:
+        if want_version.matches((maj, min)):
+            return getattr(schema, 'GET_SCHEMA_%d_%d' % (maj, min))
+
+    return schema.GET_SCHEMA_1_10
+
+
 @wsgi_wrapper.PlacementWsgify
 @microversion.version_handler('1.10')
 @util.check_accept('application/json')
@@ -243,54 +267,35 @@ def list_allocation_candidates(req):
     context = req.environ['placement.context']
     context.can(policies.LIST)
     want_version = req.environ[microversion.MICROVERSION_ENVIRON]
-    get_schema = schema.GET_SCHEMA_1_10
-    if want_version.matches((1, 31)):
-        get_schema = schema.GET_SCHEMA_1_31
-    elif want_version.matches((1, 25)):
-        get_schema = schema.GET_SCHEMA_1_25
-    elif want_version.matches((1, 21)):
-        get_schema = schema.GET_SCHEMA_1_21
-    elif want_version.matches((1, 17)):
-        get_schema = schema.GET_SCHEMA_1_17
-    elif want_version.matches((1, 16)):
-        get_schema = schema.GET_SCHEMA_1_16
+    get_schema = _get_schema(want_version)
     util.validate_query_params(req, get_schema)
 
-    requests = lib.RequestGroup.dict_from_request(req)
-    limit = req.GET.getall('limit')
-    # JSONschema has already confirmed that limit has the form
-    # of an integer.
-    if limit:
-        limit = int(limit[0])
+    rqparams = lib.RequestWideParams.from_request(req)
+    groups = lib.RequestGroup.dict_from_request(req, rqparams)
 
-    group_policy = req.GET.getall('group_policy') or None
-    # Schema ensures we get either "none" or "isolate"
-    if group_policy:
-        group_policy = group_policy[0]
-    else:
+    if not rqparams.group_policy:
         # group_policy is required if more than one numbered request group was
         # specified.
-        if len([rg for rg in requests.values() if rg.use_same_provider]) > 1:
+        if len([rg for rg in groups.values() if rg.use_same_provider]) > 1:
             raise webob.exc.HTTPBadRequest(
-                _('The "group_policy" parameter is required when specifying '
-                  'more than one "resources{N}" parameter.'))
+                'The "group_policy" parameter is required when specifying '
+                'more than one "resources{N}" parameter.')
 
     # We can't be aware of nested architecture with old microversions
     nested_aware = want_version.matches((1, 29))
 
     try:
         cands = ac_obj.AllocationCandidates.get_by_requests(
-            context, requests, limit=limit, group_policy=group_policy,
-            nested_aware=nested_aware)
+            context, groups, rqparams, nested_aware=nested_aware)
     except exception.ResourceClassNotFound as exc:
         raise webob.exc.HTTPBadRequest(
-            _('Invalid resource class in resources parameter: %(error)s') %
+            'Invalid resource class in resources parameter: %(error)s' %
             {'error': exc})
     except exception.TraitNotFound as exc:
         raise webob.exc.HTTPBadRequest(six.text_type(exc))
 
     response = req.response
-    trx_cands = _transform_allocation_candidates(cands, requests, want_version)
+    trx_cands = _transform_allocation_candidates(cands, groups, want_version)
     json_data = jsonutils.dumps(trx_cands)
     response.body = encodeutils.to_utf8(json_data)
     response.content_type = 'application/json'
