@@ -31,6 +31,7 @@ from placement import lib as pl
 from placement import microversion
 from placement.objects import resource_class as rc_obj
 from placement.objects import resource_provider as rp_obj
+from placement.tests.unit import base
 from placement import util
 
 
@@ -283,7 +284,7 @@ class TestRequireContent(testtools.TestCase):
         self.assertTrue(self.handler(req))
 
 
-class TestPlacementURLs(testtools.TestCase):
+class TestPlacementURLs(base.ContextTestCase):
 
     def setUp(self):
         super(TestPlacementURLs, self).setUp()
@@ -446,7 +447,8 @@ class TestParseQsRequestGroups(testtools.TestCase):
         mv_parsed.min_version = microversion_parse.parse_version_string(
             microversion.min_version_string())
         req.environ['placement.microversion'] = mv_parsed
-        d = pl.RequestGroup.dict_from_request(req)
+        rqparam = pl.RequestWideParams.from_request(req)
+        d = pl.RequestGroup.dict_from_request(req, rqparam)
         # Sort for easier testing
         return [d[suff] for suff in sorted(d)]
 
@@ -667,6 +669,85 @@ class TestParseQsRequestGroups(testtools.TestCase):
         self.assertRequestGroupsEqual(
             expected, self.do_parse(qs, version=(1, 24)))
 
+    def test_member_of_forbidden_aggs(self):
+        agg1_uuid = uuidsentinel.agg1
+        agg2_uuid = uuidsentinel.agg2
+        agg3_uuid = uuidsentinel.agg3
+        agg4_uuid = uuidsentinel.agg4
+        qs = ('resources=VCPU:2'
+              '&member_of=%s'
+              '&member_of=%s'
+              '&member_of=!%s'
+              '&member_of=!%s' % (
+                  agg1_uuid, agg2_uuid, agg3_uuid, agg4_uuid))
+        expected = [
+            pl.RequestGroup(
+                use_same_provider=False,
+                resources={
+                    'VCPU': 2,
+                },
+                member_of=[
+                    set([agg1_uuid]),
+                    set([agg2_uuid]),
+                ],
+                forbidden_aggs=set(
+                    [agg3_uuid, agg4_uuid]
+                ),
+            ),
+        ]
+        self.assertRequestGroupsEqual(
+            expected, self.do_parse(qs, version=(1, 32)))
+
+    def test_member_of_multiple_forbidden_aggs(self):
+        agg1_uuid = uuidsentinel.agg1
+        agg2_uuid = uuidsentinel.agg2
+        agg3_uuid = uuidsentinel.agg3
+        qs = ('resources=VCPU:2'
+              '&member_of=!in:%s,%s,%s' % (
+                  agg1_uuid, agg2_uuid, agg3_uuid))
+        expected = [
+            pl.RequestGroup(
+                use_same_provider=False,
+                resources={
+                    'VCPU': 2,
+                },
+                forbidden_aggs=set(
+                    [agg1_uuid, agg2_uuid, agg3_uuid]
+                ),
+            ),
+        ]
+        self.assertRequestGroupsEqual(
+            expected, self.do_parse(qs, version=(1, 32)))
+
+    def test_member_of_forbidden_aggs_prior_microversion(self):
+        agg1_uuid = uuidsentinel.agg1
+        agg2_uuid = uuidsentinel.agg2
+        qs = ('resources=VCPU:2'
+              '&member_of=!%s'
+              '&member_of=!%s' % (agg1_uuid, agg2_uuid))
+        self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 31))
+
+        qs = ('resources=VCPU:2'
+              '&member_of=!in:%s,%s' % (agg1_uuid, agg2_uuid))
+        self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 31))
+
+    def test_member_of_forbidden_aggs_invalid_usage(self):
+        agg1_uuid = uuidsentinel.agg1
+        agg2_uuid = uuidsentinel.agg2
+        qs = ('resources=VCPU:2'
+              '&member_of=in:%s,!%s' % (agg1_uuid, agg2_uuid))
+        self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 32))
+
+        agg1_uuid = uuidsentinel.agg1
+        agg2_uuid = uuidsentinel.agg2
+        qs = ('resources=VCPU:2'
+              '&member_of=!%s,!%s' % (agg1_uuid, agg2_uuid))
+        self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 32))
+
     def test_400_malformed_resources(self):
         # Somewhat duplicates TestNormalizeResourceQsParam.test_400*.
         qs = ('resources=VCPU:0,MEMORY_MB:4096,DISK_GB:10'
@@ -821,8 +902,52 @@ class TestParseQsRequestGroups(testtools.TestCase):
         self.assertRequestGroupsEqual(
             expected, self.do_parse(qs, version=(1, 22)))
 
+    def test_group_suffix_length_1_33(self):
+        longstring = '01234567' * 8
+        qs = 'resources_%s=CUSTOM_MAGIC:1' % longstring
+        exc = self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 33))
+        # NOTE(cdent): This error message is not what an API user would see.
+        # They would get an error during JSON schema processing.
+        self.assertIn('least one request group', str(exc))
 
-class TestPickLastModified(testtools.TestCase):
+    def test_group_suffix_character_limits_1_33(self):
+        qs = 'resources!#%=CUSTOM_MAGIC:1'
+        exc = self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 33))
+        # NOTE(cdent): This error message is not what an API user would see.
+        # They would get an error during JSON schema processing.
+        self.assertIn('least one request group', str(exc))
+
+    def test_group_suffix_character_limits_1_22(self):
+        qs = 'resources!#%=CUSTOM_MAGIC:1'
+        exc = self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 22))
+        # NOTE(cdent): This error message is not what an API user would see.
+        # They would get an error during JSON schema processing.
+        self.assertIn('least one request group', str(exc))
+
+    def test_good_suffix_1_33(self):
+        qs = ('resources_car_HOUSE_10=CUSTOM_MAGIC:1'
+              '&required_car_HOUSE_10=CUSTOM_PHYSNET1')
+        expected = [
+            pl.RequestGroup(
+                use_same_provider=True,
+                resources={
+                    'CUSTOM_MAGIC': 1,
+                },
+                required_traits={
+                    'CUSTOM_PHYSNET1',
+                }
+            ),
+        ]
+        self.assertRequestGroupsEqual(
+            expected, self.do_parse(qs, version=(1, 33)))
+        self.assertRaises(
+            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 22))
+
+
+class TestPickLastModified(base.ContextTestCase):
 
     def setUp(self):
         super(TestPickLastModified, self).setUp()
